/*  Author:		V. Alex Brennen (vab@MIT.EDU)	*/
/*  Copyright:		None				*/
/*  License:		Public Domain			*/
/*  Created:		2006.09.15			*/
/*  Last Updated:	2006.10.27			*/

#include "passgen.h"


/* This is a simple program to read some data from /dev/random or /dev/urandom and map
 * it to a set of ASCII characters.  That group of pseudorandom characters can then be
 * used as a  password.
 */

int main(int argc,char *argv[])
{
	const char version[] = "1.1.0";

	FILE 	*pseudorandom;

	char 	*buff = NULL;
	int	rslt = 0;
	int	arg = 0;
	int 	i = 0;
	int	n = 0;


	/* Init the config, then read config file */
	struct passgen_config *config = (struct passgen_config *)malloc(sizeof(struct passgen_config));
	if(config == NULL)
	{
		fprintf(stderr,"Failed to malloc config struct.\n");
		fprintf(stderr,"Perhaps, the system is out of memory?\n");
		fprintf(stderr,"Exiting...\n\n");

		return -1;
	}
	rslt = init_passgen_config(&config);
	if(rslt != 0)
	{
		fprintf(stderr,"Warning!  Failed to initialize passgen config.\n");
		fprintf(stderr,"Perhaps, the system is experiencing memory problems?\n");
		fprintf(stderr,"Exiting...\n\n");

		return -1;
	}
	rslt = read_passgen_config(&config);
	if(rslt != 0)
	{
		fprintf(stderr,"Warning!  Failed to read passgen.conf.\n");
		fprintf(stderr,"Password candidates may not match desired criteria.\n");
		fprintf(stderr,"Continuing...\n\n");
	}
	/* Options:	'-l' => Length of password
	 *		'-n' => Number of passwords to generate
	 *		'-a' => Alpha Numeric Only
	 *		'-A' => Alpha Only
	 *		'-N' => Numeric Only
	 *		'-s' => Don't use space (has distinctive sound)
	 *		'-u' => Use /dev/urandom rather than /dev/random
	 *		'-h' => Help
	 *		'-v' => Version
	 */
	if(argc > 1)
	{
		for(arg=1;arg<argc;arg++)
		{
			if(argv[arg][0] == '-')
			{
				if(argv[arg][1] == '-')
				{
					if(strstr(argv[arg],"help") != NULL)
					{
						print_help();

						return 0;
					}
					else if(strstr(argv[arg],"version") != NULL)
					{
						printf("CryptNET Password Generator Version %s\n", version);

						return 0;
					}
				}
				else if(argv[arg][1] == 'h')
				{
					print_help();

					return 0;
				}
				else if(argv[arg][1] == 'l')
				{
					/* Read in info on how many chars are wanted */
					n = arg;
					n++;
					if((argv[n] != NULL) && (isdigit(argv[n][0])))
					{
						config->num_chars = atoi(argv[n]);
					}
					else
					{
						invalid_argument();
						printf("Proper format is -l x, where x in an integer.\n");

						return 0;
					}
				}
				else if(argv[arg][1] == 'n')
				{
					/* Read in info on how many chars are wanted */
					n = arg;
					n++;
					if(((argv[n] != NULL) && isdigit(argv[n][0])))				
					{
						config->num_passwds = atoi(argv[n]);
					}
					else
					{
						invalid_argument();
						printf("Proper format is -n x, where x in an integer.\n");

						return 0;
					}
				}
				else if(argv[arg][1] == 'a')
				{
					config->passwd_type = D_ALPHA_NUMERIC;
				}
				else if(argv[arg][1] == 'A')
				{
					config->passwd_type = D_ALPHA_ONLY;
				}
				else if(argv[arg][1] == 'N')
				{
					config->passwd_type = D_NUMERIC_ONLY;
				}
				else if(argv[arg][1] == 's')
				{
					config->no_space = 1;
				}
				else if(argv[arg][1] == 'u')
				{
					config->use_urandom = 1;
				}
				else if(argv[arg][1] == 'v')
				{
					printf("CryptNET Password Generator Version %s\n", version);

					return 0;
				}
				else if(isdigit(argv[arg][0]))
				{
					continue;
				}
				else
				{
					invalid_argument();

					return 0;
				}
			}
		}
	}

	buff = (char *)malloc(config->num_chars);
	if(buff == NULL)
	{
		fprintf(stderr,"ERROR:  Failed to allocate memory for random data.\n");
		fprintf(stderr,"Perhaps, the system is out of memory?\n");
		fprintf(stderr,"Exiting...\n\n");
		if(config != NULL)
			free(config);

		return 0;
	}
	
	if(config->use_urandom == 0)
	{
		if(NULL == (pseudorandom = fopen("/dev/random","r")))
		{
			fprintf(stderr,"ERROR:  Failed to open /dev/random!\n");
			fprintf(stderr,"Exiting...\n\n");
			if(config != NULL)
				free(config);
			if(buff != NULL)
				free(buff);
	
			return 0;
		}
	}
	else
	{
		if(NULL == (pseudorandom = fopen("/dev/urandom","r")))
		{
			fprintf(stderr,"ERROR:  Failed to open /dev/urandom!\n");
			fprintf(stderr,"Exiting...\n\n");
			if(config != NULL)
				free(config);
			if(buff != NULL)
				free(buff);
	
			return 0;
		}
	}

	for(i = 0;i<config->num_passwds;i++)
	{
		memset(buff,0x00,config->num_chars);
	
		rslt = fread(buff,1,config->num_chars,pseudorandom);
		if(rslt < 0)
		{
			fprintf(stderr,"ERROR:  Read of data from specified device failed.\n");
			fclose(pseudorandom);
			if(config != NULL)
				free(config);
			if(buff != NULL)
				free(buff);
	
			return -1;
		}

		rslt = map_data_to_chars(buff,config->num_chars,config->passwd_type,config->no_space);
		if(rslt != 0)
		{
			fprintf(stderr,"ERROR:  The char mapping function returned an error.\n");
			fprintf(stderr,"Password generation failed.\n");
			fprintf(stderr,"Exiting.\n\n");
			fclose(pseudorandom);
			if(config != NULL)
				free(config);
			if(buff != NULL)
				free(buff);
	
			return 0;
		}
	}
	fclose(pseudorandom);

	if(config != NULL)
		free(config);
	if(buff != NULL)
		free(buff);

	return 0;
}


int map_data_to_chars(char *buff, int size, int type, int no_space)
{
	int lcv = 0;
	int array_size = 0;
	unsigned int tmp;
	char c = 0x00;

	char array[95] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			 'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
			 'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
			 'z','0','1','2','3','4','5','6','7','8','9','~','`','!','@','#','$',
			 '%','^','&','*','(',')','[',']','{','}','/','?','=','+','-','_','|',
			 '\\',';',':','"','\'',',','<','.','>',' '};

	char array_ns[94] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			 'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
			 'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
			 'z','0','1','2','3','4','5','6','7','8','9','~','`','!','@','#','$',
			 '%','^','&','*','(',')','[',']','{','}','/','?','=','+','-','_','|',
			 '\\',';',':','"','\'',',','<','.','>'};

	char alpha_numeric_array[62] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			 'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
			 'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
			 'z','0','1','2','3','4','5','6','7','8','9'};

	char alpha_array[52] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			 'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
			 'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
			 'z'};

	char numeric_array[10] = {'0','1','2','3','4','5','6','7','8','9'};

	/* Here array_size is array *address size*.  Arrays size is one more than array address size
	   since addresses start at zero in the C language.
	*/
	if(type == D_DEFAULT)
	{
		array_size = 94;
	}
	else if((type == D_DEFAULT) && (no_space == 1))
	{
		array_size = 93;
	}
	else if(type == D_ALPHA_NUMERIC)
	{
		array_size = 61;
	}
	else if(type == D_ALPHA_ONLY)
	{
		array_size = 51;
	}
	else if(type == D_NUMERIC_ONLY)
	{
		array_size = 9;
	}
	else
	{
		fprintf(stderr,"passgen:  Error unknown password type requested: %d.\n", type);

		return -1;
	}

	while(lcv < size)
	{
		tmp = buff[lcv] & 0x000000ff;

		if(tmp > array_size)
		{
			tmp = tmp % array_size;
		}
		if((type == D_DEFAULT) && (no_space == 1))
		{
			c = array_ns[tmp];
		}
		if(type == D_ALPHA_NUMERIC)
		{
			c = alpha_numeric_array[tmp];
		}
		else if(type == D_ALPHA_ONLY)
		{
			c = alpha_array[tmp];
		}
		else if(type == D_NUMERIC_ONLY)
		{
			c = numeric_array[tmp];
		}
		else
		{
			c = array[tmp];
		}

		/* This shouldn't be possible, but it is caught just in case. */
		if(c == '\0')
		{
			return -1;
		}
	
		printf("%c",c);

		lcv++;
	}
	printf("\n");

	return 0;
}


int print_help(void)
{
	printf("Usage: passgen\n");
	printf("\t-h\t\tThis Help Text\n");
	printf("\t-v\t\tDisplay Version Information\n");
	printf("\t-l\t\tThe Length of The Password\n");
	printf("\t-n\t\tNumber of Passwords To Provide\n");
	printf("\t-a\t\tAlpha Numeric Characters Only\n");
	printf("\t-A\t\tAlphabetic Characters Only\n");
	printf("\t-N\t\tNumeric Characters Only\n");
	printf("\t-s\t\tExclude The Space Character\n");
	printf("\t-u\t\tUse /dev/urandom rather than /dev/random \n");
	printf("\t\t\t(Will speed password generation)\n");
	printf("\t--help\t\tThis Help Text\n");
	printf("\t--version\tDisplay Version Information\n");
	printf("\n");
	printf("Persisting alternate default configuration values may be set in passgen.conf.\n");
	printf("For for information: `man passgen.conf`\n");
	printf("\n");

	return 0;
}


int invalid_argument(void)
{
	fprintf(stderr,"passgen: Invalid arument given\n\n");
	
	print_help();

	return 0;
}
